@@ -87,6 +87,11 @@ class Group(GroupIoMixin):
         self.sum += int(value)
         self.log_prod += log(factorial(value))
 
+    def add_repeated_value(self, shared, value, count):
+        self.count += count
+        self.sum += int(count * value)
+        self.log_prod += count * log(factorial(value))
+
     def remove_value(self, shared, value):
         self.count -= 1
         self.sum -= int(value)
