@@ -136,6 +136,12 @@ class Group(GroupIoMixin):
         self.mean += delta / self.count
         self.count_times_variance += delta * (value - self.mean)
 
+    def add_repeated_value(self, shared, value, count):
+        self.count += count
+        delta = count * value - self.mean
+        self.mean += delta / self.count
+        self.count_times_variance += delta * (value - self.mean)
+
     def remove_value(self, shared, value):
         total = self.mean * self.count
         delta = value - self.mean
