@@ -42,13 +42,20 @@ struct DirichletProcessDiscrete
 {
 
 typedef DirichletProcessDiscrete Model;
-typedef uint32_t count_t;
+typedef int count_t;
 typedef uint32_t Value;
 struct Group;
 struct Scorer;
 struct Sampler;
-struct VectorizedScorer;
-typedef GroupScorerMixture<VectorizedScorer> Mixture;
+struct MixtureDataScorer;
+struct MixtureValueScorer;
+typedef MixtureSlave<Model, MixtureDataScorer> SmallMixture;
+typedef MixtureSlave<Model, MixtureDataScorer, MixtureValueScorer> FastMixture;
+typedef FastMixture Mixture;
+
+static constexpr Value OTHER () { return 0xFFFFFFFFU; }
+static constexpr float MIN_BETA () { return 1e-6f; }
+
 
 struct Shared : SharedMixin<Model>
 {
@@ -58,14 +65,13 @@ struct Shared : SharedMixin<Model>
     Sparse_<Value, float> betas;
     SparseCounter<Value, count_t> counts;
 
-    static constexpr Value OTHER () { return 0xFFFFFFFFU; }
-
     void add_value (const Value & value, rng_t & rng)
     {
         DIST_ASSERT1(value != OTHER(), "cannot add OTHER");
         if (DIST_UNLIKELY(counts.add(value) == 1)) {
-            float beta = beta0 * sample_beta(rng, 1.f, gamma);
-            beta0 = std::max(0.f, beta0 - beta);
+            DIST_ASSERT(beta0 > 0, "cannot add any more values");
+            float beta = beta0 * sample_beta_safe(rng, 1.f, gamma, MIN_BETA());
+            beta0 = std::max(MIN_BETA(), beta0 - beta);
             betas.add(value, beta);
         }
     }
@@ -74,7 +80,7 @@ struct Shared : SharedMixin<Model>
     {
         DIST_ASSERT1(value != OTHER(), "cannot remove OTHER");
         if (DIST_UNLIKELY(counts.remove(value) == 0)) {
-            beta0 += betas.pop(value);
+            beta0 = std::min(1.f, beta0 + betas.pop(value));
         }
     }
 
@@ -193,17 +199,28 @@ struct Group : GroupMixin<Model>
             const Value & value,
             rng_t &)
     {
-        DIST_ASSERT1(value != shared.OTHER(), "cannot add OTHER");
+        DIST_ASSERT1(value != OTHER(), "cannot add OTHER");
         DIST_ASSERT1(shared.betas.contains(value), "unknown value: " << value);
         counts.add(value);
     }
 
+    void add_repeated_value (
+            const Shared & shared,
+            const Value & value,
+            const int & count,
+            rng_t &)
+    {
+        DIST_ASSERT1(value != OTHER(), "cannot add OTHER");
+        DIST_ASSERT1(shared.betas.contains(value), "unknown value: " << value);
+        counts.add(value, count);
+    }
+
     void remove_value (
             const Shared & shared,
             const Value & value,
             rng_t &)
     {
-        DIST_ASSERT1(value != shared.OTHER(), "cannot remove OTHER");
+        DIST_ASSERT1(value != OTHER(), "cannot remove OTHER");
         DIST_ASSERT1(shared.betas.contains(value), "unknown value: " << value);
         counts.remove(value);
     }
@@ -222,7 +239,7 @@ struct Group : GroupMixin<Model>
             rng_t &) const
     {
         float alpha = shared.alpha;
-        float numer = (value == shared.OTHER())
+        float numer = (value == OTHER())
                     ? alpha * shared.beta0
                     : alpha * shared.betas.get(value) + counts.get_count(value);
         float denom = alpha + counts.get_total();
@@ -239,9 +256,6 @@ struct Group : GroupMixin<Model>
         float score = 0;
         for (auto & i : counts) {
             Value value = i.first;
-            DIST_ASSERT3(
-                shared.betas.contains(value),
-                "shared.betas is missing value: " << value);
             float prior_i = alpha * shared.betas.get(value);
             score += fast_lgamma(prior_i + i.second)
                    - fast_lgamma(prior_i);
@@ -260,6 +274,18 @@ struct Group : GroupMixin<Model>
         sampler.init(shared, *this, rng);
         return sampler.eval(shared, rng);
     }
+
+    void validate (const Shared & shared) const
+    {
+        for (const auto & i : counts) {
+            if (auto group_count = i.second) {
+                auto shared_count = shared.counts.get_count(i.first);
+                DIST_ASSERT(
+                    shared_count,
+                    "shared_count = 0 but group_count = " << group_count);
+            }
+        }
+    }
 };
 
 struct Sampler
@@ -284,7 +310,7 @@ struct Sampler
             probs.push_back(beta * alpha + group.counts.get_count(value));
         }
         if (shared.beta0 > 0) {
-            values.push_back(shared.OTHER());
+            values.push_back(OTHER());
             probs.push_back(shared.beta0 * alpha);
         }
 
@@ -313,7 +339,7 @@ struct Scorer
 
         const size_t total = group.counts.get_total();
         const float beta_scale = shared.alpha / (shared.alpha + total);
-        scores.add(shared.OTHER(), beta_scale * shared.beta0);
+        scores.add(OTHER(), beta_scale * shared.beta0);
         for (auto & i : shared.betas) {
             scores.add(i.first, i.second * beta_scale);
         }
@@ -338,149 +364,11 @@ struct Scorer
     }
 };
 
-struct VectorizedScorer : VectorizedScorerMixin<Model>
+struct MixtureDataScorer : MixtureSlaveDataScorerMixin<Model, MixtureDataScorer>
 {
-    void resize (const Shared & shared, size_t size)
-    {
-        scores_shift_.resize(size);
-        for (const auto & i : shared.betas) {
-            Value value = i.first;
-            scores_.get_or_add(value).resize(size);
-        }
-        if (scores_.size() != shared.betas.size()) {
-            for (auto i = scores_.begin(); i != scores_.end();) {
-                if (DIST_UNLIKELY(not shared.betas.contains(i->first))) {
-                    scores_.unsafe_erase(i++);
-                } else {
-                    ++i;
-                }
-            }
-        }
-    }
-
-    void add_group (const Shared & shared, rng_t &)
-    {
-        const float alpha = shared.alpha;
-        for (auto & i : scores_) {
-            i.second.packed_add(fast_log(alpha * shared.betas.get(i.first)));
-        }
-        scores_shift_.packed_add(fast_log(alpha));
-    }
-
-    void remove_group (const Shared &, size_t groupid)
-    {
-        for (auto & i : scores_) {
-            i.second.packed_remove(groupid);
-        }
-        scores_shift_.packed_remove(groupid);
-    }
-
-    void update_group (
-            const Shared & shared,
-            size_t groupid,
-            const Group & group,
-            rng_t &)
-    {
-        const float alpha = shared.alpha;
-        for (auto & i : scores_) {
-            Value value = i.first;
-            count_t count = group.counts.get_count(value);
-            i.second[groupid] =
-                fast_log(alpha * shared.betas.get(value) + count);
-        }
-        count_t count_sum = group.counts.get_total();
-        scores_shift_[groupid] = fast_log(alpha + count_sum);
-    }
-
-    void add_value (
-            const Shared & shared,
-            size_t groupid,
-            const Group & group,
-            const Value & value,
-            rng_t &)
-    {
-        DIST_ASSERT1(value != shared.OTHER(), "cannot add OTHER");
-        if (DIST_UNLIKELY(not scores_.contains(value))) {
-            const size_t group_count = scores_shift_.size();
-            const float beta = shared.alpha * shared.betas.get(value);
-            scores_.add(value).resize(group_count, fast_log(beta));
-        }
-
-        _update_group_value(shared, groupid, group, value);
-    }
-
-    void remove_value (
-            const Shared & shared,
-            size_t groupid,
-            const Group & group,
-            const Value & value,
-            rng_t &)
-    {
-        DIST_ASSERT1(value != shared.OTHER(), "cannot remove OTHER");
-        if (DIST_UNLIKELY(not shared.betas.contains(value))) {
-            scores_.remove(value);
-        }
-
-        _update_group_value(shared, groupid, group, value);
-    }
-
-    void update_all (
-            const Shared & shared,
-            const MixtureSlave<Shared> & slave,
-            rng_t &)
-    {
-        const size_t group_count = slave.groups().size();
-        const float alpha = shared.alpha;
-
-        for (auto & i : scores_) {
-            Value value = i.first;
-            VectorFloat & scores = i.second;
-            const float beta = shared.betas.get(value);
-            for (size_t groupid = 0; groupid < group_count; ++groupid) {
-                auto count = slave.groups(groupid).counts.get_count(value);
-                scores[groupid] = alpha * beta + count;
-            }
-            vector_log(group_count, scores.data());
-        }
-
-        for (size_t groupid = 0; groupid < group_count; ++groupid) {
-            auto total = slave.groups(groupid).counts.get_total();
-            scores_shift_[groupid] = alpha + total;
-        }
-        vector_log(group_count, scores_shift_.data());
-    }
-
-    void score_value (
-            const Shared & shared,
-            const Value & value,
-            VectorFloat & scores_accum,
-            rng_t &) const
-    {
-        if (DIST_LIKELY(scores_.contains(value))) {
-
-            vector_add_subtract(
-                scores_accum.size(),
-                scores_accum.data(),
-                scores_.get(value).data(),
-                scores_shift_.data());
-
-        } else {
-
-            float beta = (value == shared.OTHER())
-                       ? shared.beta0
-                       : shared.betas.get(value);
-            float score = fast_log(shared.alpha * beta);
-            vector_add_subtract(
-                scores_accum.size(),
-                scores_accum.data(),
-                score,
-                scores_shift_.data());
-        }
-    }
-
     float score_data (
             const Shared & shared,
-            const MixtureSlave<Shared> & slave,
+            const std::vector<Group> & groups,
             rng_t &) const
     {
         const float alpha = shared.alpha;
@@ -492,7 +380,7 @@ struct VectorizedScorer : VectorizedScorerMixin<Model>
         const float shared_total = fast_lgamma(alpha);
 
         float score = 0;
-        for (const auto & group : slave.groups()) {
+        for (const auto & group : groups) {
             if (group.counts.get_total()) {
                 for (auto & i : group.counts) {
                     Value value = i.first;
@@ -507,33 +395,216 @@ struct VectorizedScorer : VectorizedScorerMixin<Model>
 
         return score;
     }
+};
 
-    void score_data_grid (
-            const std::vector<Shared> & shareds,
-            const MixtureSlave<Shared> & slave,
-            AlignedFloats scores_out,
-            rng_t & rng) const
+struct MixtureValueScorer : MixtureSlaveValueScorerMixin<Model>
+{
+    void resize (const Shared & shared, size_t size)
     {
-        slave.score_data_grid(shareds, scores_out, rng);
+        scores_shift_.resize(size);
+        for (const auto & i : shared.betas) {
+            Value value = i.first;
+            auto & entry = scores_.get_or_add(value);
+            entry.ref_count = 1;
+            entry.scores.resize(size);
+        }
+        if (scores_.size() != shared.betas.size()) {
+            for (auto i = scores_.begin(); i != scores_.end();) {
+                if (DIST_UNLIKELY(not shared.betas.contains(i->first))) {
+                    scores_.unsafe_erase(i++);
+                } else {
+                    ++i;
+                }
+            }
+        }
+
+        _validate(shared, size);
+    }
+
+    void add_group (const Shared & shared, rng_t &)
+    {
+        const float alpha = shared.alpha;
+        for (auto & i : scores_) {
+            i.second.scores.packed_add(
+                fast_log(alpha * shared.betas.get(i.first)));
+        }
+        scores_shift_.packed_add(fast_log(alpha));
+    }
+
+    void remove_group (const Shared &, size_t groupid)
+    {
+        for (auto & i : scores_) {
+            i.second.scores.packed_remove(groupid);
+        }
+        scores_shift_.packed_remove(groupid);
+    }
+
+    void update_group (
+            const Shared & shared,
+            size_t groupid,
+            const Group & group,
+            rng_t &)
+    {
+        const float alpha = shared.alpha;
+        for (auto & i : scores_) {
+            Value value = i.first;
+            auto & entry = i.second;
+            count_t count = group.counts.get_count(value);
+            entry.scores[groupid] =
+                fast_log(alpha * shared.betas.get(value) + count);
+        }
+        scores_shift_[groupid] = fast_log(alpha + group.counts.get_total());
+    }
+
+    void add_value (
+            const Shared & shared,
+            size_t groupid,
+            const Group & group,
+            const Value & value,
+            rng_t &)
+    {
+        DIST_ASSERT1(value != OTHER(), "cannot add OTHER");
+        auto & entry = scores_.get_or_add(value);
+        ++entry.ref_count;
+        if (DIST_UNLIKELY(entry.ref_count == 1)) {
+            const size_t group_count = scores_shift_.size();
+            const float beta = shared.alpha * shared.betas.get(value);
+            entry.scores.resize(group_count, fast_log(beta));
+        }
+        entry.scores[groupid] = fast_log(
+            shared.alpha * shared.betas.get(value) +
+            group.counts.get_count(value));
+        scores_shift_[groupid] = fast_log(
+            shared.alpha + group.counts.get_total());
+    }
+
+    void remove_value (
+            const Shared & shared,
+            size_t groupid,
+            const Group & group,
+            const Value & value,
+            rng_t &)
+    {
+        DIST_ASSERT1(value != OTHER(), "cannot remove OTHER");
+        auto & entry = scores_.get(value);
+        --entry.ref_count;
+        if (DIST_UNLIKELY(entry.ref_count == 0)) {
+            scores_.remove(value);
+        } else {
+            entry.scores[groupid] = fast_log(
+                shared.alpha * shared.betas.get(value) +
+                group.counts.get_count(value));
+        }
+        scores_shift_[groupid] = fast_log(
+            shared.alpha + group.counts.get_total());
+    }
+
+    void update_all (
+            const Shared & shared,
+            const std::vector<Group> & groups,
+            rng_t &)
+    {
+        _validate(shared, groups.size());
+        const size_t group_count = groups.size();
+        const float alpha = shared.alpha;
+
+        for (auto & i : scores_) {
+            Value value = i.first;
+            auto & entry = i.second;
+            entry.ref_count = 0;
+            const float beta = shared.betas.get(value);
+            for (size_t groupid = 0; groupid < group_count; ++groupid) {
+                auto count = groups[groupid].counts.get_count(value);
+                entry.ref_count += count;
+                entry.scores[groupid] = alpha * beta + count;
+            }
+            vector_log(group_count, entry.scores.data());
+        }
+
+        for (size_t groupid = 0; groupid < group_count; ++groupid) {
+            auto total = groups[groupid].counts.get_total();
+            scores_shift_[groupid] = alpha + total;
+        }
+        vector_log(group_count, scores_shift_.data());
+    }
+
+    float score_value_group (
+            const Shared & shared,
+            const std::vector<Group> & groups,
+            size_t groupid,
+            const Value & value,
+            rng_t &) const
+    {
+        _validate(shared, groups.size());
+
+        if (DIST_LIKELY(scores_.contains(value))) {
+            return scores_.get(value).scores[groupid] - scores_shift_[groupid];
+        } else {
+            float beta = (value == OTHER())
+                       ? shared.beta0
+                       : shared.betas.get(value);
+            return fast_log(shared.alpha * beta) - scores_shift_[groupid];
+        }
+    }
+
+    void score_value (
+            const Shared & shared,
+            const std::vector<Group> & groups,
+            const Value & value,
+            AlignedFloats scores_accum,
+            rng_t &) const
+    {
+        _validate(shared, groups.size());
+
+        if (DIST_LIKELY(scores_.contains(value))) {
+
+            vector_add_subtract(
+                scores_accum.size(),
+                scores_accum.data(),
+                scores_.get(value).scores.data(),
+                scores_shift_.data());
+
+        } else {
+
+            float beta = (value == OTHER())
+                       ? shared.beta0
+                       : shared.betas.get(value);
+            float score = fast_log(shared.alpha * beta);
+            vector_add_subtract(
+                scores_accum.size(),
+                scores_accum.data(),
+                score,
+                scores_shift_.data());
+        }
     }
 
 private:
 
-    void _update_group_value (
-            const Shared & shared,
-            size_t groupid,
-            const Group & group,
-            const Value & value)
+    void _validate (const Shared & shared, size_t group_count) const
     {
-        const float alpha = shared.alpha;
-        count_t count = group.counts.get_count(value);
-        count_t count_sum = group.counts.get_total();
-        scores_.get(value)[groupid] =
-            fast_log(alpha * shared.betas.get(value) + count);
-        scores_shift_[groupid] = fast_log(alpha + count_sum);
+        if (DIST_DEBUG_LEVEL >= 2) {
+            DIST_ASSERT_LE(scores_.size(), shared.betas.size());
+            DIST_ASSERT_EQ(scores_shift_.size(), group_count);
+        }
+        if (DIST_DEBUG_LEVEL >= 3) {
+            for (const auto & i : scores_) {
+                const Value & value = i.first;
+                const auto & entry = i.second;
+                DIST_ASSERT(
+                    shared.betas.contains(value),
+                    "missing value: " << value);
+                DIST_ASSERT_EQ(entry.scores.size(), group_count);
+            }
+        }
     }
 
-    Sparse_<Value, VectorFloat> scores_;
+    struct CountAndScores
+    {
+        uint32_t ref_count;
+        VectorFloat scores;
+        CountAndScores () : ref_count(0), scores() {}
+    };
+    Sparse_<Value, CountAndScores> scores_;
     VectorFloat scores_shift_;
 };
 
