@@ -44,8 +44,11 @@ typedef float Value;
 struct Group;
 struct Scorer;
 struct Sampler;
-struct VectorizedScorer;
-typedef GroupScorerMixture<VectorizedScorer> Mixture;
+struct MixtureDataScorer;
+struct MixtureValueScorer;
+typedef MixtureSlave<Model, MixtureDataScorer> SmallMixture;
+typedef MixtureSlave<Model, MixtureDataScorer, MixtureValueScorer> FastMixture;
+typedef FastMixture Mixture;
 
 
 struct Shared : SharedMixin<Model>
@@ -101,7 +104,7 @@ struct Shared : SharedMixin<Model>
 
 struct Group : GroupMixin<Model>
 {
-    uint32_t count;
+    int count;
     float mean;
     float count_times_variance;
 
@@ -141,6 +144,18 @@ struct Group : GroupMixin<Model>
         count_times_variance += delta * (value - mean);
     }
 
+    void add_repeated_value (
+            const Shared &,
+            const Value & value,
+            const int & count,
+            rng_t &)
+    {
+        this->count += count;
+        float delta = count * value - mean;
+        mean += delta / this->count;
+        count_times_variance += delta * (value - mean);
+    }
+
     void remove_value (
             const Shared &,
             const Value & value,
@@ -168,7 +183,7 @@ struct Group : GroupMixin<Model>
             const Group & source,
             rng_t &)
     {
-        uint32_t total_count = count + source.count;
+        auto total_count = count + source.count;
         float delta = source.mean - mean;
         float source_part = float(source.count) / total_count;
         float cross_part = count * source_part;
@@ -268,7 +283,36 @@ struct Scorer
     }
 };
 
-struct VectorizedScorer : VectorizedScorerMixin<Model>
+struct MixtureDataScorer : MixtureSlaveDataScorerMixin<Model, MixtureDataScorer>
+{
+    float score_data (
+            const Shared & shared,
+            const std::vector<Group> & groups,
+            rng_t &) const
+    {
+        const float nu_part = fast_lgamma(0.5f * shared.nu);
+        const float kappa_part = 0.5f * fast_log(shared.kappa);
+        const float sigmasq_part =
+            0.5f * shared.nu * fast_log(shared.nu * shared.sigmasq);
+        const float log_pi = 1.1447298858493991f;
+
+        float score = 0;
+        for (const auto & group : groups) {
+            if (group.count) {
+                Shared post = shared.plus_group(group);
+                score += fast_lgamma(0.5f * post.nu) - nu_part;
+                score += kappa_part - 0.5f * fast_log(post.kappa);
+                score += sigmasq_part
+                       - 0.5f * post.nu * fast_log(post.nu * post.sigmasq);
+                score += -0.5f * log_pi * group.count;
+            }
+        }
+
+        return score;
+    }
+};
+
+struct MixtureValueScorer : MixtureSlaveValueScorerMixin<Model>
 {
     void resize(const Shared &, size_t size)
     {
@@ -334,57 +378,34 @@ struct VectorizedScorer : VectorizedScorerMixin<Model>
 
     void update_all (
             const Shared & shared,
-            const MixtureSlave<Shared> & slave,
+            const std::vector<Group> & groups,
             rng_t & rng)
     {
-        const size_t group_count = slave.groups().size();
+        const size_t group_count = groups.size();
         for (size_t groupid = 0; groupid < group_count; ++groupid) {
-            update_group(shared, groupid, slave.groups()[groupid], rng);
+            update_group(shared, groupid, groups[groupid], rng);
         }
     }
 
+    float score_value_group (
+            const Shared &,
+            const std::vector<Group> &,
+            size_t groupid,
+            const Value & value,
+            rng_t &) const
+    {
+        float temp = 1.f + precision_[groupid] * sqr(value - mean_[groupid]);
+        return score_[groupid] + log_coeff_[groupid] * fast_log(temp);
+    }
+
     // not thread safe
     void score_value (
             const Shared &,
+            const std::vector<Group> &,
             const Value & value,
             AlignedFloats scores_accum,
             rng_t &) const;
 
-    float score_data (
-            const Shared & shared,
-            const MixtureSlave<Shared> & slave,
-            rng_t &) const
-    {
-        const float nu_part = fast_lgamma(0.5f * shared.nu);
-        const float kappa_part = 0.5f * fast_log(shared.kappa);
-        const float sigmasq_part =
-            0.5f * shared.nu * fast_log(shared.nu * shared.sigmasq);
-        const float log_pi = 1.1447298858493991f;
-
-        float score = 0;
-        for (const auto & group : slave.groups()) {
-            if (group.count) {
-                Shared post = shared.plus_group(group);
-                score += fast_lgamma(0.5f * post.nu) - nu_part;
-                score += kappa_part - 0.5f * fast_log(post.kappa);
-                score += sigmasq_part
-                       - 0.5f * post.nu * fast_log(post.nu * post.sigmasq);
-                score += -0.5f * log_pi * group.count;
-            }
-        }
-
-        return score;
-    }
-
-    void score_data_grid (
-            const std::vector<Shared> & shareds,
-            const MixtureSlave<Shared> & slave,
-            AlignedFloats scores_out,
-            rng_t & rng) const
-    {
-        slave.score_data_grid(shareds, scores_out, rng);
-    }
-
 private:
 
     VectorFloat score_;
